/*	From FreeBSD: */
/*
 *      tcpblast - test and estimate TCP throuput
 *
 *      Daniel Karrenberg   <dfk@nic.eu.net>
 */

/*
 *	Changes: Rafal Maszkowski <rzm@icm.edu.pl>
 *
 *	ftp://6bone-gw.6bone.pl/pub/blast/README
 */

char verstr[30]="FreeBSD + rzm ", *vdate=VERSION;

#include <sys/types.h>
#include <getopt.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>	/* for Solaris */
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>	/* fcntl() */
#include <signal.h>	/* signal() */

#if defined(__sunos__) || defined(__osf1__)
#define strerror(x)     atoi(x)
char * gettext(x) char *x; { return x; }
#endif

/* to fit time difference in one variable we are scaling s*(1000*USCALE)+us/(1000/USCALE) */
#define	USCALE		100
#define DEFBLKSIZE	(1024)
#define MAXBLKSIZE	(32*1024)

/* options to make printresult() arguments list shorter */
#define PRRES_VERBMASK	7
#define	PRRES_SCROLL	8
#define	PRRES_RD	16
#define	PRRES_WR	32
#define	PRRES_SHOW	64

/* circular buffer element containing time and written/read bytes */
typedef struct {
	long	t, b;
} cbelem;

struct	sockaddr_in sock_in;
struct	servent *sp;
struct	hostent *host;
struct	timeval t0;
struct	timezone tz;
char 	wbuff[MAXBLKSIZE], rbuff[MAXBLKSIZE], *ind, port[30]="9";
long	readall=0, writeall=0;
unsigned long	delay=0;
int 	f, bufsize=0, tcp=0, udp=0, randomb=0, blksize=DEFBLKSIZE, setbufsize=-1,
	dots=1, continuous=1, experimental=0, verbosity=0, lastnwb=1,
	lastnrb=1, reading=0, writing=1, comparing=0, scroll=0, maxwrblk=300,
	maxrdblk=0, maxtime=0, lport=0 /* any */	;

void
usage(char *name) {	
	fprintf(stderr, "\n");
	fprintf(stderr, "Usage: %s [options] destination[:port]\n\n", name);
	fprintf(stderr, "tcpblast/udpblast is a simple tool for probing network and estimating its\n");
	fprintf(stderr, "throughput. By default it sends %d of data to specified destination host.\n", maxwrblk);
	fprintf(stderr, "Each block has %d B by default.\n\n", blksize);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "-a               send random data\n");
	fprintf(stderr, "-b BUF_SIZE      socket buf size (default: %d == %s), with `-' to be substracted from results\n", setbufsize, setbufsize==-1 ? "don't change" : "change");
	fprintf(stderr, "-c BLOCKS        change default (%d) number of blocks, range: 1..9999\n", maxwrblk);
	fprintf(stderr, "-d DOTFREQ       print dot every DOTFREQ blocks, disables cont. speed disp. \n");
/*	fprintf(stderr, "-f FILE          send FILE instead of generated data\n") ;		*/
	fprintf(stderr, "-h, --help       this help\n");
	fprintf(stderr, "-i, --delay DELAY  write delay in microseconds (EXPERIMENTAL)\n");
	fprintf(stderr, "-l, --last BLKS  show also speed for last BLKS blocks\n");
	fprintf(stderr, "    --nwrite     do not write, use e.g. with chargen port\n");
	fprintf(stderr, "-m               results for every block in separate line\n");
	fprintf(stderr, "-o               turn off default continuous speed displaying\n");
	fprintf(stderr, "-p PORT          bind this local PORT\n");
	fprintf(stderr, "-r, --read       read data returned to us, switches default port to echo\n");
	fprintf(stderr, "-s BLOCK_SIZE    block size (default %d bytes)\n", blksize);
	fprintf(stderr, "-t MAXTIME       limit time to MAXTIME s\n");
	fprintf(stderr, "    --tcp        use TCP (%s)\n", ind[0]=='t' ? "default" : "default if named tcpblast" );
	fprintf(stderr, "    --udp        use UDP (%s)\n", ind[0]=='u' ? "default" : "default if named udpblast" );
	fprintf(stderr, "-v, --verbosity  verbosity, default 0, maximum 3\n");
	fprintf(stderr, "-V, --version    version\n");
	fprintf(stderr, "destination      host name or address\n");
	fprintf(stderr, "port             use port #/name xyz instead of default %s\n", port);
	fprintf(stderr, "\n");
	fprintf(stderr, "%s version: %s; max. time limit: %7.1f h\n", name, verstr, 4.294967295e9/(1000*USCALE));
	exit(0);
}

void usage_small(char *name) {	
	fprintf(stderr, "type %s --help for help\n", name);
	exit(0);
}

/* randomize the buffer */
void
randbuff(int blksize) {
	int i;
	for (i=0; i<blksize; i++) {
		wbuff[i]=rand() % 256;
	}
}

unsigned long
gettdiff(struct timeval t0) {
	struct timeval t1;
	struct timezone t1z;

	if (gettimeofday(&t1, &t1z) < 0)
	{
		perror("tcp/udpblast time");
		exit(11);
	}
	return (t1.tv_sec-t0.tv_sec)*(1000*USCALE) + (t1.tv_usec-t0.tv_usec)/(1000/USCALE);
}

/* shoud be ready for options = PRRES_RD|PRRES_WR */

/*
	t0			start time
	writeall/readall	current # of B written/read
	bufsize			buffer size substracted form the results
	lastn[wr]b		if >0 we additionally count speed fo last n blocks
	options			PRRES_VERBMASK - 3 bits, PRRES_SCROLL, PRRES_RD, PRRES_WR, PRRES_SHOW
	global: max{wr|rd}blk	loop limir for written/read blocks
*/

void
printresult(struct timeval t0, long writeall, long readall, int bufsize, int lastnwb, int lastnrb, int options) {
	static int circrbufset = 0, circwbufset = 0, cbrsize, cbwsize, cbrptr, cbwptr;
	static cbelem *circrbuf = NULL, *circwbuf = NULL;
	unsigned long wtime, rtime, dwtime, drtime, pwtime, prtime;
	int rd, wr, vrb, show;

	wtime = rtime = gettdiff(t0);

	rd = options&PRRES_RD; wr = options&PRRES_WR;
	vrb = options&PRRES_VERBMASK; show = options&PRRES_SHOW;

	if (show && !(options&PRRES_SCROLL)) printf("\r");

	/* possible future expansion: reset the table if # of KBs changes (or shrink/extend in next version) */
	if (wr && circwbufset && (lastnwb != cbwsize)) {
		circwbufset = 0;
		if (circwbuf) free(circwbuf);
	}
	if (rd && circrbufset && (lastnrb != cbrsize)) {
		circrbufset = 0;
		if (circrbuf) free(circrbuf);
	}

	if (wr && (lastnwb>0) && circwbufset) {
		/* we are not advancing the pointer if no new bytes were found;
		   comparing with b+blksize handles special case when we finish blocked
		   and # of bytes is uneven, we can avoid using a new time after long
		   blocking, we will still use the last # of bytes when calculating speed */
		if (writeall >= circwbuf[cbwptr % cbwsize].b + blksize) {
			cbwptr++;
			circwbuf[cbwptr % cbwsize].t = wtime;
			circwbuf[cbwptr % cbwsize].b = writeall;
		} else {
			/* instead of dissolving the rate we are using 'stale' style */
			wtime = circwbuf[cbwptr % cbwsize].t;
		}
		pwtime = circwbuf[(cbwptr + 1) % cbwsize].t;
		if (pwtime >= wtime) { dwtime = 0; } else { dwtime = wtime - pwtime; };
	} else {
		circwbuf = malloc(sizeof(cbelem)*lastnwb);
		if (circwbuf == NULL) {
			perror("malloc: no memory");
		} else {
			bzero(circwbuf, sizeof(cbelem)*lastnwb);
			cbwsize = lastnwb;
			cbwptr = 0;
			circwbufset = 1;
		}
	}

	if (rd && (lastnrb>0) && circrbufset) {
		/* see comment for writing */
		if (readall >= circrbuf[cbrptr % cbrsize].b + blksize) {
			cbrptr++;
			circrbuf[cbrptr % cbrsize].t = rtime;
			circrbuf[cbrptr % cbrsize].b = readall;
		} else {
			/* instead of dissolving the rate we are using 'stale' style */
			rtime = circrbuf[cbrptr % cbrsize].t;
		}
		prtime = circrbuf[(cbrptr + 1) % cbrsize].t;
		if (prtime >= rtime) { drtime = 0; } else { drtime = rtime - prtime; };
	} else {
		circrbuf = malloc(sizeof(cbelem)*lastnrb);
		if (circrbuf == NULL) {
			perror("malloc: no memory");
		} else {
			bzero(circrbuf, sizeof(cbelem)*lastnrb);
			cbrsize = lastnrb;
			cbrptr = 0;
			circrbufset = 1;
		}
	}

	if (show) {
		if (wr) printf("Written");
		if (rd) printf("/Read");
		printf(" ");
		if (wr) printf("%5ld", writeall/1024);
		if (rd) printf("/%ld", readall/1024);
		printf(" KB");
		if (vrb >= 1) {
			printf("  in ");
			if (wr) printf("%.1f", (double)wtime/USCALE);
			if (rd) printf("/%.1f", (double)rtime/USCALE);
			printf(" ms");
		}
		if (vrb >= 3) {
			printf("   ");
			if (wr) printf("%.1f", ((double)writeall-bufsize)/wtime*8000*USCALE);
			if (rd) printf("/%.1f", ((double)readall-bufsize)/rtime*8000*USCALE);
			printf(" b/s");
		}
		if (vrb >= 2) {
			printf("   ");
			if (wr) printf("%.1f", ((double)writeall-bufsize)/wtime*1000*USCALE);
			if (rd) printf("/%.1f", (double)(readall-bufsize)/rtime*USCALE*1000);
			printf(" B/s");
		}
		printf("   ");
		if (wr) printf("%.2f", ((double)writeall-bufsize) / (double) (wtime*(1024.0/USCALE)) * 1000 );
		if (rd) printf("/%.2f", ((double)readall-bufsize) / (double) (rtime*(1024.0/USCALE)) * 1000 );
		printf(" KB/s");
		if ( (lastnwb>1) || (lastnrb>1) ) {
			long dw, dr;
			if (wr) dw = writeall-circwbuf[(cbwptr + 1) % cbwsize].b;
			if (rd) dr = readall -circrbuf[(cbrptr + 1) % cbrsize].b;
			printf("  (last ");
			if (wr) printf("%ld", dw/1024);
			if (rd) printf("/%ld", dr/1024);
			printf(" KB:");
			printf(" ");
			if (wr) printf("%.1f", (double)dw /dwtime*USCALE*1000 );
			if (rd) printf("/%.1f", (double)dr /drtime*USCALE*1000);
			printf(" B/s)");
		}
		printf("     ");
		if (options&PRRES_SCROLL) printf("\n");
		fflush(stdout);
	}
}

/*
 * finish -- (netkit-base-0.10-37, ping.c)
 *      Print out statistics, and give up.
 */

void
finish() {
        /* (void)signal(SIGINT, SIG_IGN);	*/

/* printf("\n"); */
	if (dots!=0) printf("\n");
	if (continuous) printf("\r");
	printresult(t0, writeall, readall, 0, lastnwb, lastnrb, verbosity|scroll|reading*PRRES_RD|writing*PRRES_WR|PRRES_SHOW);
	if (experimental && (bufsize>0) && (maxwrblk*blksize>bufsize)) {
		printf("\nExperimentally taking into account %d B socket buffer:\n", bufsize);
		printresult(t0, writeall, readall, bufsize, lastnwb, lastnrb,
			verbosity|scroll|reading*PRRES_RD|writing*PRRES_WR|PRRES_SHOW);
	}
	printf("\n");
	exit(0);
}

int
main(int argc, char **argv) {
	long	written1, read1, tmpw, tmpr, wrblk, rdblk, flags;
	char	hostname[100], *colon;
	struct servent *service;

	strcat(verstr, vdate);

	/* non-random data - is modem compressing it? */
	bzero(wbuff, MAXBLKSIZE);
	/*	memset(wbuff, 'a', MAXBLKSIZE); 	oops, tests */

	/* find first letter in the name - usage() needs it */
	ind=rindex(argv[0], '/');
	if (ind==NULL) ind=argv[0]; else ind++;

	while (1) {
		static const struct option long_options[] = {
			{ "delay",	required_argument, NULL, 'i' },
			{ "help",	no_argument,	NULL,	'h' },
			{ "nwrite",	no_argument,	NULL,	1 },
			{ "read",	no_argument,	NULL,	'r' },
			{ "tcp",	no_argument,	NULL,	2 },
			{ "udp",	no_argument,	NULL,	3 },
			{ "version",	no_argument,	NULL,	'V' },
			{ "verbosity",	no_argument,	NULL,	'v' },
			{ NULL, 0, NULL, 0 }
		};
		char	optchar;
		if ((optchar = getopt_long (argc, argv, "ab:c:d:hi:l:mop:rs:t:vV", long_options, NULL)) == -1) break;
		switch (optchar) {
			case 1: writing=0;							break;
			case 2: tcp=1;								break;
			case 3: udp=1;								break;
			case 'a': randomb=1;							break;
			case 'b':
				setbufsize=atoi(optarg);	/* negative BUFSIZE means substracting it from results */
				if (setbufsize < 0) bufsize = setbufsize = abs(setbufsize);
				break;
			case 'c': maxwrblk=abs(atoi(optarg));					break;
			case 'd': dots=abs(atoi(optarg)); continuous=0;				break;
			case 'f': printf("not yet done\n"); exit(1);				break;
			case 'h': usage(argv[0]);						break;
			case 'i': delay=abs(atoi(optarg));					break;
			case 'l': lastnwb=lastnrb=abs(atoi(optarg)); 				break; /* split later */
			case 'm': scroll=PRRES_SCROLL;						break;
			case 'o': continuous=0;							break;
			case 'p': lport=atoi(optarg);
				if ((lport<0) || (lport>65535)) {
					printf("Local port must be in 0..65535 range\n");
					exit(1);
				}								break;
			case 'r': reading=1; strcpy(port,"7");					break;
			case 's': blksize=abs(atoi(optarg));					break;
			case 't': maxtime=abs(atoi(optarg));					break;
			case 'v': verbosity++;							break;
			case 'V': printf("%s version: %s\n", argv[0], verstr);	return 0;	break;
			case '?': exit(1);
			default: ;
		}
	}

/* correctness */
	if (tcp && udp) {
		printf("cannot use both TCP and UDP\n");
		usage_small(argv[0]);
		exit(2);
	}

	/* if neither -t nor -u is chosen use first character of the program name */
	if ( (tcp==0) && (udp==0) && (ind[0]=='t') ) tcp=1;
	if ( (tcp==0) && (udp==0) && (ind[0]=='u') ) udp=1;

	if (!tcp && !udp) {
		printf("must use either TCP or UDP\n");
		usage_small(argv[0]);
		exit(3);
	}

	if (continuous) dots=0;

	/* after options processing we need one arg left */
	if (argc - optind != 1) {
		if (argc - optind == 0) printf("give a hostname[:port]\n");
		usage_small(argv[0]);
		exit(4);
	}
	strncpy(hostname, argv[optind], sizeof(hostname)-1);

        if (maxwrblk<=0 || maxwrblk>=10000) {
		fprintf(stderr, "%s: %d BLOCKS outside of range 1..9999 \n", argv[0], maxwrblk);
		exit(5);
	}
	if (!maxrdblk) maxrdblk = maxwrblk;
	if (!reading)  maxrdblk = 0;
	if (!writing)  maxwrblk = 0;

	bzero((char *)&sock_in, sizeof (sock_in));
	sock_in.sin_family = AF_INET;
	sock_in.sin_port = htons(lport);
	if (tcp) f = socket(AF_INET, SOCK_STREAM, 0);
	else     f = socket(AF_INET, SOCK_DGRAM,  0);
	if (f < 0) {
		perror("tcp/udpblast: socket");
		exit(6);
	}

	{ int size=sizeof(int); 	/* getsockopt() should know how much space we have */
		/* get/setsockopt doesn't return any error really for SO_SNDBUF,
		   at least on Linux; it limits the buffer to [2048..65536]
		   (131070 for 2.1 (?) but you can manipulate with /proc/sys/net/core/wmem_max ) */
		if (getsockopt(f, SOL_SOCKET, SO_SNDBUF, (void *)&bufsize, &size)==-1)
			printf("%s getsockopt: %s", argv[0], strerror(errno));
		printf("read SO_SNDBUF = %d\n", bufsize);
		if (setbufsize!=-1) {
			if (setsockopt(f, SOL_SOCKET, SO_SNDBUF, (void *)&setbufsize, sizeof(setbufsize))==-1)
				printf("%s getsockopt: %s", argv[0], strerror(errno));
			if (getsockopt(f, SOL_SOCKET, SO_SNDBUF, (void *)&bufsize, &size)==-1) /* size value's been set b4 */
				printf("%s getsockopt: %s", argv[0], strerror(errno));
			printf("set  SO_SNDBUF = %d\n", bufsize);
		}
	}

	if (bind(f, (struct sockaddr*)&sock_in, sizeof (sock_in)) < 0) {
		perror("tcp/udpblast: bind");
		exit(7);
	}

	if ((colon=index(hostname, ':')) != NULL) {
		*colon = '\0';
		strncpy(port, colon+1, sizeof(port)-1);
	}

	host = gethostbyname(hostname);
	if (host) {
		sock_in.sin_family = host->h_addrtype;
		bcopy(host->h_addr, &sock_in.sin_addr, host->h_length);
	} else {
		sock_in.sin_family = AF_INET;
		sock_in.sin_addr.s_addr = inet_addr(argv[optind]);
		if (sock_in.sin_addr.s_addr == -1) {
			fprintf(stderr, "%s: %s unknown host\n", argv[0], argv[optind]);
			exit(8);
		}
	}

	/* port # or name can be used */
	service = getservbyname(port, tcp ? "tcp" : "udp");
	if (service==NULL) sock_in.sin_port = htons(abs(atoi(port)));
	else               sock_in.sin_port = service->s_port;

	printf("Sending %s %s data to %s:%s using %d B blocks.\n",
		randomb ? "random":"non-random", tcp ? "TCP":"UDP", hostname, port, blksize);

	if (connect(f, (struct sockaddr*)&sock_in, sizeof(sock_in)) <0) {
		perror("tcp/udpblast connect");
		exit(9);
	}

	if (gettimeofday(&t0, &tz) < 0) {
		perror("tcp/udpblast time");
		exit(10);
	}
	if (randomb) srand(t0.tv_usec);		/* more less... */

	/* single process, all non-blocking writes/reads */
	flags=fcntl(f, F_GETFL);
	fcntl(f, F_SETFL, flags|O_NONBLOCK);

	if (comparing) {
		/* allocate a table of pointers to all possible blocks */
	}

	signal(SIGINT, finish);		/* ^C stops */

	written1 = read1 = wrblk = rdblk = 0;
	do {
		if (writing && (wrblk<maxwrblk)) {
			tmpw = written1;
			if (delay) usleep(delay);
			written1 = write(f, wbuff+written1, (size_t)blksize-written1);
			if (written1 != -1) {
				writeall += written1;
				written1 += tmpw;
			} else { written1 = tmpw; }
			if (written1 >= blksize) {
				written1 -= blksize; wrblk++;
				printresult(t0, writeall, readall, experimental ? bufsize : 0,
					lastnwb, lastnrb, verbosity|scroll|PRRES_WR|reading*PRRES_RD
					|writing*PRRES_WR|continuous*PRRES_SHOW);
				if ( (dots) && ( (dots==1) || (wrblk%dots==1) ) ) write(1, ".", 1);
				if (randomb) {
					randbuff(blksize);
					/* alloc space, copy */
					if (comparing) {
					}
				}
			}
		}

		if (reading && (rdblk<maxrdblk)) {
			tmpr = read1;		/* saving to cumulate after reading */
			read1=read(f, rbuff+read1, (size_t)blksize-read1);
			if (read1 != -1) {
				readall += read1;
				read1 += tmpr;
			} else { read1 = tmpr; }
			if (read1 >= blksize) {
				read1 -= blksize; rdblk++;
				printresult(t0, writeall, readall, experimental ? bufsize : 0,
					lastnwb, lastnrb, verbosity|scroll|PRRES_RD|reading*PRRES_RD
					|writing*PRRES_WR|continuous*PRRES_SHOW);
				if ( (dots) && ( (dots==1) || (rdblk%dots==1) ) ) write(1, "\b \b", 1);
				if (comparing) {
					if (randomb) {
					} else {
					}
				}
			}
		}

		if (maxtime && ((gettdiff(t0)/(1000*USCALE)) > maxtime)) break;
	} while ( (wrblk<maxwrblk) || (rdblk<maxrdblk) );

	finish();
	return(1);	/* for the compiler mostly */
}
